/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  the contents of this file will be loaded into the Mnenhy.folderstore.* context!
// |  ie. foo() is available as goMnenhy.folderstore.foo()
// +-
// |  used user_prefs:
// |    extensions.mnenhy.log.folderstore       bool    debug log
// |    extensions.mnenhy.folderstore.noprefs   bool    don't save user_prefs
// |    extensions.mnenhy.folderstore.account   bool    always fall back to account data (if any)
// |    extensions.mnenhy.folderstore.onenter   string  code to be executed when entering folder
// |    extensions.mnenhy.folderstore.onexit    string  code to be executed when leaving folder
// +-


// |
// |  constants
// |

// files
const gksFolderstoreRDF    = "mnenhy-folderstore.rdf";
const gksDefFolderstoreRDF = "chrome://mnenhy/content/folderstore/" + gksFolderstoreRDF;
// prefs
const gksPrefDebug         = "extensions.mnenhy.log.folderstore";
const gksPrefNoPrefs       = "extensions.mnenhy.folderstore.noprefs";
const gksPrefAccount       = "extensions.mnenhy.folderstore.account";
const gksPrefOnEnter       = "extensions.mnenhy.folderstore.onenter";
const gksPrefOnExit        = "extensions.mnenhy.folderstore.onexit";
// persistence
const gksFolderObserver    = "mail:setupToolbarItems";
const gksLastDataFolder    = "mnenhy:folderstore:LastData";
const gksPersist           = "persist";
// import/export
const gkoFilePicker        = Components.interfaces.nsIFilePicker;
const gksFilePickerDefault = "MnenhyWatchedPrefs";
const gksFilePickerDefExt  = ".js";
const gksPrefNodePrefix    = "//mnenhy=";
const gksNewLine           = "\r\n";
const gksPrefLinePrefix    = 'user_pref("';
const gksPrefLineMid       = '", ';
const gksPrefLinePostfix   = ');' + gksNewLine;

const gksStringTablePrefix = "mnenhy";
const gkxPrefLine          = new RegExp("(?:user_)?pref\\((" + gksStringTablePrefix + "\\d+),("
                             + gksStringTablePrefix+ "\\d+|true|false|null|\\d+)\\);?");
const gkxPrefLines         = new RegExp(gkxPrefLine.source, "g");


// |
// |  pseudo constants
// |

// text properties
var gksConfirmInherit     = null;
var gksFilePickerTitleIm  = null;
var gksFilePickerTitleEx  = null;
var gksFilePickerJSTitle  = null;
var gksFilePickerJSFile   = null;
var gksFilePickerAllTitle = null;
var gksFilePickerAllFile  = null;


// other
var goMnenhy          = null;   // global Mnenhy object
var gkxTestNS         = null;   // /^ + goMnenhy.ksMnenhyNamespace/
var rdf               = null;   // mnenhy-folderstore.rdf
var gkoTypePrefix     = {};     // converts pref type constants to text
var gkiTypePrefixLen  = 0;      // length of each type constant's text
var goPrefCache       = {};     // pref cache: elements = {type: iType, value: sValue};

var dd                = null;   // debug dump function


// |
// |  variables
// |

// current and clicked folder URIs for context menu and config dialog
var gsCurrentURI      = null;
var gsClickedURI      = null;
var gsCurrentDataURI  = null;
var gsClickedDataURI  = null;

// |
// |  folder-dependent pref and threadtree-persistance storage
// |

// we get notified of folder changes...
function MnenhyFolderChange(aoWindow, asTopic, asNewFolderURI)
{ // |this| is the observer object!
  const koThis = aoWindow.goMnenhy.folderstore;
  koThis.dd('MnenhyFolderChange('+aoWindow+','+asTopic+','+asNewFolderURI+')');

  // if the folder storage isn't fully loaded, try again in 10ms
  if (!koThis.rdf.mbDataSourceLoaded)
  {
    aoWindow.setTimeout(koThis.MnenhyFolderChange, 10, aoWindow, asTopic, asNewFolderURI);
    return;
  }

  // virtual folders send the URI of the base folder here, not their own :-/
  if (("gCurrentVirtualFolderUri" in aoWindow) && aoWindow.gCurrentVirtualFolderUri)
    asNewFolderURI = aoWindow.gCurrentVirtualFolderUri;

  // we do not need to memorize which real *folder* we came from, all we need to know is
  // whether the *data folder* (containing the prefs and persists) changed!
  var oDataFolder = {ksOld: "", ksNew: ""};
  if (koThis.IsChange(aoWindow, asTopic, asNewFolderURI, oDataFolder))
    koThis.DoFolderChange(aoWindow, oDataFolder.ksOld, oDataFolder.ksNew);
}


// process folderchange regardless of necessity
function DoFolderChange(aoWindow, asOldFolderURI, asNewFolderURI)
{ // Part A: store the current (old) folder-dependent data
  this.SaveFolderData(aoWindow, asOldFolderURI);
  // Part B: read the new folder-dependent data
  this.LoadFolderData(aoWindow, asNewFolderURI);
}


// Part A: store the current (old) folder-dependent data
//   (A.1) store current (old) threadpane persist data in RDF
//   (A.2) execute the onExit handler
//   (A.3) store current (old) pref values in RDF
//   (A.4) restore pref values from goPrefCache
//   (A.5) clear goPrefCache
function SaveFolderData(aoWindow, asDataFolder)
{ // (A.1) store current (old) threadpane persist data in RDF
  // we need a non-empty target URI when persisting the threadpane
  this.dd("SaveFolderData(" + aoWindow + "," + asDataFolder + ")");
  this.PersistThreadpane(aoWindow, true, asDataFolder);

  // (A.2) execute some code in the context of the current window
  const ksEvalOnExit = goMnenhy.GetPref(this.gksPrefOnExit, goMnenhy.kiPrefString);
  if (ksEvalOnExit)
  {
    try
    { // TB 1.0.6 doesn't like eval(ksEvalOnExit, aoWindow); it complains about
      // "EvalError: function eval must be called directly, and not by way of a function of another name."
      if ("eval" in aoWindow)
      { // But using |aoWindow.eval(ksEvalOnExit, aoWindow);| will provoke the same message,
        // just as a warning, when in strict mode. Unfortunately, we need to use a deprecated
        // language feature ('with') to keep the JavaScript console clean...
        with (aoWindow)
        {
          eval(ksEvalOnExit, aoWindow);
        }
      }
      else
        // if all else fails, try it the old way
        eval(ksEvalOnExit, aoWindow);
    }
    catch(oErr)
    {
      goMnenhy.Log("SaveFolderData: eval of '"+ksEvalOnExit+"' failed:\n"+oErr);
    }
  }

  const kbStorePrefs = !goMnenhy.GetPref(this.gksPrefNoPrefs, goMnenhy.kiPrefBool);
  if (kbStorePrefs)
  { // Hey, we can process the next two steps in just one loop! \o/
    // (A.3) store current (old) pref values in RDF
    // (A.4) restore pref values from goPrefCache
    var sPref, iType, oValue;
    for (sPref in this.goPrefCache)
    { // assume that the names in the goPrefCache are exactly the same as in the RDF...
      // (and if they aren't, it can't be our fault! ;-) )
      // (A.3) store current (old) pref values in RDF
      iType = this.goPrefCache[sPref].type;
      if (iType == goMnenhy.kiPrefInvalid)
        iType = goMnenhy.kiPrefString;  // assume string type for unknown prefs
      oValue = this.gkoTypePrefix[iType] + goMnenhy.GetPref(sPref, iType);
      const ksPref = goMnenhy.ksMnenhyNamespace + encodeURIComponent(sPref);  // add namespace
      this.dd("SaveFolderData: store pref in RDF: " + ksPref + "=" + oValue);
      this.rdf.ChangeItemValue(oValue, ksPref, asDataFolder);
      // (A.4) restore pref values from goPrefCache
      iType  = this.goPrefCache[sPref].type;
      oValue = this.goPrefCache[sPref].value;
      this.dd("SaveFolderPrefs: restoring pref: "+sPref+"="+oValue+" ("+iType+")");
      goMnenhy.SetPref(sPref, iType, oValue);
    }

    // (A.5) clear goPrefCache
    this.goPrefCache = {}; // clear cache
  }
}


// Part B: read the new folder-dependent data
//   (B.1) read new threadpane persist data from RDF
//   (B.2) cache current pref values in goPrefCache
//   (B.3) read new pref values from RDF
//   (B.4) execute the onEnter handler
//   (B.5) memorize the new old data folder
function LoadFolderData(aoWindow, asDataFolder)
{ // (B.1) read new threadpane persist data from RDF
  // we need a non-empty target URI when persisting the threadpane
  this.dd("LoadFolderData(" + aoWindow + "," + asDataFolder + ")");
  this.PersistThreadpane(aoWindow, false, asDataFolder);

  const kbStorePrefs = !goMnenhy.GetPref(this.gksPrefNoPrefs, goMnenhy.kiPrefBool);
  if (kbStorePrefs)
  { // Hey again, we can process the next two steps in just one loop also! \o/
    // But we need a local callback function for this...
    if (this.rdf.ExistNode(asDataFolder))
    { // read the prefs
      // (B.2) cache current pref values in goPrefCache
      // (B.3) read new pref values from RDF
      this.rdf.GetAllArcsOut(asDataFolder, this.ReadAndCacheFolderPrefs, this);
    }
  }

  // (B.4) execute some code in the context of the current window
  const ksEvalOnEnter = goMnenhy.GetPref(this.gksPrefOnEnter, goMnenhy.kiPrefString);
  if (ksEvalOnEnter)
  {
    try
    { // TB 1.0.6 doesn't like eval(ksEvalOnEnter, aoWindow); it complains about
      // "EvalError: function eval must be called directly, and not by way of a function of another name."
      if ("eval" in aoWindow)
      { // But using |aoWindow.eval(ksEvalOnEnter, aoWindow);| will provoke the same message,
        // just as a warning, when in strict mode. Unfortunately, we need to use a deprecated
        // language feature ('with') to keep the JavaScript console clean...
        with (aoWindow)
        {
          eval(ksEvalOnEnter, aoWindow);
        }
      }
      else
        // if all else fails, try it the old way
        eval(ksEvalOnEnter, aoWindow);
    }
    catch(oErr)
    {
      goMnenhy.Log("LoadFolderData: eval of '"+ksEvalOnEnter+"' failed:\n"+oErr);
    }
  }

  // (B.5) memorize the new old data folder
  this.SetLastFolder(aoWindow, asDataFolder);
}


// callback for (B.2) and (B.3) in LoadFolderData above
function ReadAndCacheFolderPrefs(aoResourceIgnored, aoItem, aoValue, aoThis)
{
  if (    (aoItem  instanceof goMnenhy.rdf.gkIRDFResource)
       && (aoValue instanceof goMnenhy.rdf.gkIRDFLiteral)
       && (!aoItem.Value.search(aoThis.gkxTestNS))
     )
  {
    aoThis.dd('ReadAndCacheFolderPrefs: '+aoItem.Value+'='+aoValue.Value);
    // the type of the pref *must* be stored in the RDF value!
    const koMnenhy = aoThis.goMnenhy;
    const ksPref   = decodeURIComponent(aoItem.Value.replace(aoThis.gkxTestNS, ""));
    var sValue = aoValue.Value.substr(0, aoThis.gkiTypePrefixLen);
    var iType  = koMnenhy.kiPrefInvalid;
    for (var i in aoThis.gkoTypePrefix)
    {
      if (sValue == aoThis.gkoTypePrefix[i])
      { // found it!
        iType  = Number(i);
        sValue = aoValue.Value.substr(aoThis.gkiTypePrefixLen);
        break;
      }
    }
    aoThis.dd('ReadAndCacheFolderPrefs: pref='+ksPref+', type='+iType+', val='+sValue);
    if (iType != koMnenhy.kiPrefInvalid)
    { // (B.2) cache current pref values in goPrefCache
      aoThis.goPrefCache[ksPref] = {type: iType, value: koMnenhy.GetPref(ksPref, iType)};
      // (B.3) set new pref values from RDF
      koMnenhy.SetPref(ksPref, iType, sValue);
    }
  }
  return true; // loop over all outgoing arcs
}


// cleanup on shutdown
function MnenhyFolderUnload(aoEventIgnored)
{ // |this| is the chrome window!
  const koThis = goMnenhy.folderstore;
  koThis.dd("MnenhyFolderUnload("+koThis.GetGlobalURI(this)+")");
  // restore the global threadtree layout
  koThis.MnenhyFolderChange(this, koThis.gksFolderObserver, koThis.GetGlobalURI(this));
  // flush the folderstorage
  koThis.rdf.FlushDataSource();
}


// Loop over all treecol elements of this window's threadtree and call ProcessElement for them.
// (abSave will be passed to ProcessElement, denoting save/read)
function PersistThreadpane(aoWindow, abSave, asFolderURI)
{
  if (!aoWindow)
    aoWindow = window;
  const koThis = aoWindow.goMnenhy.folderstore;
  koThis.dd("PersistThreadpane("+aoWindow+","+abSave+","+asFolderURI+")");
  if (!asFolderURI)
    return;
  const koUserData = {save: abSave, uri: asFolderURI, cx: koThis};
  koThis.WalkThreadTreeColumns(aoWindow,
                               koThis.ProcessElement,
                               koUserData);
  // we need to persist the thread tree element "lastfoldersent" also
  koThis.ProcessElement(aoWindow.GetThreadTree(), koUserData);
}


// Loop over all the persistent attributes of the element and read or save them.
function ProcessElement(aoElement, aoUserData)
{
  const koThis = aoUserData.cx;
  koThis.dd("ProcessElement("+aoElement+","+aoUserData+")"); // don't use uneval on aoUserData!
  const ksFolderURI = aoUserData.uri;
  if (!aoElement || !ksFolderURI)
    return;

  // get the list of persistent elements
  const krsPersist = aoElement.getAttribute(koThis.gksPersist).split(" ");
  var sAttr, sValue;
  for (var i in krsPersist)
  {
    sAttr = krsPersist[i];
    if (sAttr)
    {
      if (aoUserData.save)
      { // save persistent data
        sValue = aoElement.getAttribute(sAttr);
        koThis.WriteAttribute(ksFolderURI, aoElement.id, sAttr, sValue);
      }
      else
      { // load persistent data
        sValue = koThis.ReadAttribute(ksFolderURI, aoElement.id, sAttr);
        if (sValue)
          aoElement.setAttribute(sAttr, sValue);
        else
          aoElement.removeAttribute(sAttr);
      }
    } // if attr
  } // for attr
}


// Get the attribute value for attribute asAttrName of the
// persist element with id [asFolderURI]#[asColID].
// Valid return values are strings, including empty ones; |null| denotes an error!
function ReadAttribute(asFolderURI, asID, asAttrName)
{
  if (!asFolderURI || !asID || !asAttrName)
    return null;

  // get the persist object
  const koDS = this.rdf;
  const ksID = asFolderURI + "#" + asID;
  // does this node exist in this container?
  if (!koDS.ExistNode(ksID, asFolderURI))
    return null;

  const ksValue = koDS.GetItemValue(asAttrName, ksID);
  this.dd('ReadAttribute('+asFolderURI+"#"+asID+': '+asAttrName+'='+ksValue+')');
  return ksValue;
}


// set the attribute value asValue of attribute asAttrName for the
// persist element with id [asFolderURI]#[asID]
function WriteAttribute(asFolderURI, asID, asAttrName, asValue)
{
  if (!asFolderURI || !asID || !asAttrName)
    return;

  this.dd('WriteAttribute('+asFolderURI+"#"+asID+': '+asAttrName+'='+asValue+')');
  const koDS = this.rdf;
  const ksID = asFolderURI + "#" + asID;
  // check parent container this.persistBaseID
  if (!koDS.ExistNode(asFolderURI))
    koDS.AddContainer(asFolderURI);
  // make sure the data node is stored in the parent container
  if (koDS.ExistNode(asFolderURI))
  { // have parent
    if (!koDS.ExistNode(ksID, asFolderURI))
    { // no data node in parent container
      koDS.AddNode(ksID, asFolderURI);
    }
    if (koDS.ExistNode(ksID, asFolderURI))
    { // have data node in parent container
      koDS.ChangeItemValue(asValue, asAttrName, ksID);
    }
  } // have container
}


// |
// |  context menu and config dialog functions
// |

// copy the parent data to here, hence breaking inheritence
function CopyFromParent(aoWindow)
{
  this.dd('CopyFromParent('+aoWindow+')');
  // save current data
  var sDataFolder = this.GetDataFolder(aoWindow, this.gsCurrentURI, true);
  this.SaveFolderData(aoWindow, sDataFolder);
  // copy data
  const ksParentDataURI = this.GetDataFolder(aoWindow, this.gsClickedURI, false);
  this.CopyFolderData(aoWindow, ksParentDataURI, this.gsClickedURI);
  // always refresh now to show changes
  sDataFolder = this.GetDataFolder(aoWindow, this.gsCurrentURI, true);
  this.LoadFolderData(aoWindow, sDataFolder);
}


// copy data from clicked to current
function CopyFromHere(aoWindow)
{ // copy from here to current
  this.dd('CopyFromParent('+aoWindow+')');
  // save current data
  var sDataFolder = this.GetDataFolder(aoWindow, this.gsCurrentURI, true);
  this.SaveFolderData(aoWindow, sDataFolder);
  // copy data
  this.CopyFolderData(aoWindow, this.gsClickedURI, this.gsCurrentURI);
  // refresh now to show changes
  sDataFolder = this.GetDataFolder(aoWindow, this.gsCurrentURI, true);
  this.LoadFolderData(aoWindow, sDataFolder);
}


function CopyFolderData(aoWindow, asSourceURI, asTargetURI)
{ // copy folder data from source to target
  // if asSourceURI is empty, copy global data
  // (during runtime, the global data is stored in our folderstore.rdf, too!)
  this.dd("CopyFolderData("+aoWindow+","+asSourceURI+","+asTargetURI+")+");
  if (asTargetURI)
  {
    if (!asSourceURI)
    {
      asSourceURI = this.GetGlobalURI(aoWindow);
      this.dd("CopyFolderData: using global source="+asSourceURI);
    }
    if (asSourceURI && asSourceURI != asTargetURI)
    {
      var sOldNode, sNewNode, oUserData;
      // create target container
      if (!this.rdf.ExistNode(asTargetURI) || !this.rdf.InitContainer(asTargetURI))
        this.rdf.AddContainer(asTargetURI)

      const koUserData = {source: asSourceURI, target: asTargetURI};
      // loop over the threadtree's persistent columns and copy the respective data
      // (the thread tree needs to have the MnenhyPersistTree binding!)
      this.WalkThreadTreeColumns(aoWindow, this.CopyNodeItems, koUserData);
      // we need to copy the thread tree element "lastfoldersent" also
      this.CopyNodeItems(aoWindow.GetThreadTree(), koUserData);
      // now copy the pref data of the parent node
      this.CopyNodeItems(null, koUserData);
    }
  }
  this.dd("CopyFolderData-");
}


// copy all item/value arcs from asSourceNode to asTargetNode
function CopyNodeItems(aoElement, aoUserData)
{
  const koThis = goMnenhy.folderstore;
  koThis.dd("CopyNodeItems("+aoElement+","+uneval(aoUserData)+")");
  // if aoElement is given, add its ID
  var sSourceNode = aoUserData.source;
  var sTargetNode = aoUserData.target;
  if (aoElement)
  {
    sSourceNode += "#" + aoElement.id;
    sTargetNode += "#" + aoElement.id;
  }
  // create target node if it doesn't exist yet
  if (!koThis.rdf.ExistNode(sTargetNode))
    koThis.rdf.AddNode(sTargetNode, aoUserData.target);
  if (koThis.rdf.ExistNode(sTargetNode))
  {
    const koUserData = {rdf: koThis.rdf, uri: sTargetNode};
    koThis.rdf.GetAllArcsOut(sSourceNode, koThis.CopyRDFArc, koUserData);
  }
}
function CopyRDFArc(aoResourceIgnored, aoItem, aoValue, aoUserData)
{
  if (    (aoItem  instanceof goMnenhy.rdf.gkIRDFResource)
       && (aoValue instanceof goMnenhy.rdf.gkIRDFLiteral))
  { // copy the data
    aoUserData.rdf.ChangeItemValue(aoValue.Value, aoItem.Value, aoUserData.uri);
  }
  return true; // continue with next element
}


// |
// |  helper functions
// |

function SetLastFolder(aoWindow, asURI)
{
  aoWindow[aoWindow.goMnenhy.folderstore.gksLastDataFolder] = asURI;
}

// Check if the announced folder change is relevant for us.
// If it is, aoData will contain the old and new folder URI
function IsChange(aoWindow, asTopic, asNewFolderURI, aoData)
{
  this.dd('IsChange('+aoWindow+','+asTopic+','+asNewFolderURI+')');
  // only act on _folder_ URI change, not for news-message: URI etc.
  if (    asTopic != this.gksFolderObserver
       || (asNewFolderURI && !asNewFolderURI.search(/^.+?-message:/))
     )
    return false;

  if (!aoData)
    aoData = {};
  if (this.gksLastDataFolder in aoWindow)
  { // we have old data to check against
    aoData.ksOld = aoWindow[this.gksLastDataFolder];
  }
  else
  { // no old data means that we are at startup:
    // store the threadpane layout from localstore.rdf in our rdf
    // and set that as the old data folder
    aoData.ksOld = GetGlobalURI(aoWindow);
    this.PersistThreadpane(aoWindow, true, aoData.ksOld);
    // we need to clean up on window shutdown
    aoWindow.addEventListener("unload", this.MnenhyFolderUnload, false);
    // don't add listener twice
    this.SetLastFolder(aoWindow, aoData.ksOld);
  }
  aoData.ksNew = this.GetDataFolder(aoWindow, asNewFolderURI, true);
  var bChange = aoData.ksOld != aoData.ksNew;
  this.dd('IsChange='+bChange+'; aoData='+uneval(aoData));
  return bChange;
}


// walk up the folder hierarchy until we find some data ;-)
// if all else fails, we get the global URI
function GetDataFolder(aoWindow, asBaseURI, abCheckBase)
{
  this.dd("GetDataFolder: ("+aoWindow+","+asBaseURI+","+abCheckBase+")");
  if (goMnenhy.GetPref(this.gksPrefAccount, goMnenhy.kiPrefBool))
  { // we must not check chrome:// servers - they do not exist ;-)
    if (asBaseURI && asBaseURI.substr(0,7) != "chrome:")
    {
      const koServer = aoWindow.GetServer(asBaseURI);
      if (koServer instanceof Components.interfaces.nsIMsgIncomingServer)
        asBaseURI = koServer.serverURI;
      else
      { // if we cannot get the server for this account, use global
        asBaseURI = "";
        this.dd("GetDataFolder: no server!");
      }
    }
    this.dd("GetDataFolder: using account="+asBaseURI);
  }
  // we already have a data folder
  if (abCheckBase && this.rdf.ExistNode(asBaseURI))
    return asBaseURI;

  var rsPath = asBaseURI.split('/'), sParentURI, bNoNode = true;
  while (rsPath.length && bNoNode)
  {
    rsPath.pop();
    sParentURI = rsPath.join('/');
    bNoNode = !this.rdf.ExistNode(sParentURI, null, true);
  }
  return bNoNode ? this.GetGlobalURI(aoWindow) : sParentURI;
}


// get the URI of the window
function GetGlobalURI(aoWindow)
{
  return aoWindow.document.location.href;
}


// Loop over all treecol elements of the window's threadtree, calling
// afCallback for each, passing aoUserData through
function WalkThreadTreeColumns(aoWindow, afCallback, aoUserData)
{
  const koThis = aoWindow.goMnenhy.folderstore;
  koThis.dd("WalkThreadTreeColumns("+aoWindow+","+(afCallback?afCallback.name:null)+","+aoUserData+")");
  if (!aoWindow || !afCallback)
    return;

  // get the treecols element
  var oPersistCols = aoWindow.GetThreadTree().firstChild;
  while (oPersistCols && (oPersistCols.nodeName != "treecols"))
    oPersistCols = oPersistCols.nextSibling;

  // create a treewalker to loop over all the treecol elements
  const koTreeWalker = aoWindow.document.createTreeWalker(oPersistCols,
                                                          NodeFilter.SHOW_ELEMENT,
                                                          koThis.TreeColFilter,
                                                          false);
  if (koTreeWalker)
  { // process all tree columns
    for (var oCol = koTreeWalker.nextNode(); oCol; oCol = koTreeWalker.nextNode())
      afCallback(oCol, aoUserData);
  }
}
function TreeColFilter(aoNode)
{
  if (aoNode.nodeName == "treecol")
    return NodeFilter.FILTER_ACCEPT;
  return NodeFilter.FILTER_REJECT;
}


// |
// |  Import/Export of watched user preferences
// |

function GetImExFileName(asFolderURI)
{
  return gksFilePickerDefault + (asFolderURI ? "." + asFolderURI : "") + gksFilePickerDefExt;
}

function ImportPrefs(asFolderURI /* = "" */)
{ // Read watched preferences from a file and store them in our folderstore.rdf.
  // We expect the input file to be written by our export function
  // and/or the Mozilla preferences service:
  //
  // * Pseudo-BNF for Mozilla preferences files, taken from prefread.cpp:
  // * ------------------------------------------------------------------
  // * function      = LJUNK function-name JUNK function-args
  // * function-name = "user_pref" | "pref"
  // * function-args = "(" JUNK pref-name JUNK "," JUNK pref-value JUNK ")" JUNK ";"
  // * pref-name     = quoted-string
  // * pref-value    = quoted-string | "true" | "false" | integer-value
  // * JUNK          = *(WS | comment-block | comment-line)
  // * LJUNK         = *(WS | comment-block | comment-line | bcomment-line)
  // * WS            = SP | HT | LF | VT | FF | CR
  // * SP            = <US-ASCII SP, space (32)>
  // * HT            = <US-ASCII HT, horizontal-tab (9)>
  // * LF            = <US-ASCII LF, linefeed (10)>
  // * VT            = <US-ASCII HT, vertical-tab (11)>
  // * FF            = <US-ASCII FF, form-feed (12)>
  // * CR            = <US-ASCII CR, carriage return (13)>
  // * comment-block = <C/C++ style comment block>
  // * comment-line  = <C++ style comment line>
  // * bcomment-line = <bourne-shell style comment line>
  //
  // As an extension, we accept the literal value "null" to delete prefs on import!
  //
  // A typical input file consists of an anonymous section, followed by several
  // named Mnenhy export sections. The anonymous section will be added to every
  // named section that is processed!
  // If no asFolderURI is given, all named sections are imported.
  // If asFolderURI is given, only the respective section is read.
  this.dd("ImportPrefs("+asFolderURI+")");
  var oFP = Components.classes[goMnenhy.ksCID_FilePicker].createInstance(gkoFilePicker);
  oFP.init(window, gksFilePickerTitleIm, gkoFilePicker.modeOpen);
  oFP.appendFilter(gksFilePickerJSTitle,  gksFilePickerJSFile );
  oFP.appendFilter(gksFilePickerAllTitle, gksFilePickerAllFile);
  oFP.displayDirectory = goMnenhy.m_koDirUser.clone();  // default = user profile
  oFP.defaultString    = GetImExFileName(asFolderURI);
  const kiResult = oFP.show();
  if (kiResult == gkoFilePicker.returnOK && oFP.file.exists())
  {
    goMnenhy.SetWaitCursor(window);
    var sImport = goMnenhy.ReadFile(oFP.file);
    this.dd("ImportPrefs: sImport=\n"+sImport);
    // remove junk
    const kxPrefNodePrefix = new RegExp("^" + gksPrefNodePrefix);
    var rImport  = [], iPos, xh, sh;
    var osImport = {}, ix = 0;  // string table
    while (sImport.length)
    { // lookout for the next comment, string or whitespace
      iPos = sImport.search(/\/\/|\/\*|#|"|'|\s+/);
      if (iPos < 0)
      { // keep the entire remainder
        rImport.push(sImport);
        sImport = "";
      }
      else
      { // keep all non-problematic chars
        rImport.push(sImport.substr(0, iPos));
        sImport = sImport.substr(iPos);
        switch (sImport[0])
        {
          case '/':
            switch (sImport[1])
            {
              case '/': // ignore remainder of line -
                        // if it isn't a Mnenhy export section mark
                iPos = sImport.search(/\n/) + 1;
                if (!sImport.search(kxPrefNodePrefix))
                { // keep comment
                  rImport.push(sImport.substring(0, iPos));
                }
                break;

              case '*': // ignore stream comment
                iPos = sImport.search(/\*\//) + 2;
            }
            break;

          case '#': // ignore remainder of line
            iPos = sImport.search(/\n/) + 1;
            break;

          case '"': // keep whole string (search next unescaped ")
            xh = /(?:[^\\](?:\\\\)*)"/.exec(sImport);
            if (xh)
            {
              iPos = xh.index + xh[0].length;
              sh = gksStringTablePrefix + ix++;
              osImport[sh] = sImport.substring(1, iPos - 1);
              rImport.push(sh);
            }
            else
              sImport = ""; // error; ignore remainder
            break;

          case "'": // keep whole string (search next unescaped ')
            xh = /(?:[^\\](?:\\\\)*)'/.exec(sImport);
            if (xh)
            {
              iPos = xh.index + xh[0].length;
              sh = gksStringTablePrefix + ix++;
              osImport[sh] = sImport.substring(1, iPos - 1);
              rImport.push(sh);
            }
            else
              sImport = ""; // error; ignore remainder
            break;

          default:  // ignore whitespace
            iPos = sImport.search(/[^\s]/);
            if (iPos < 0)
              iPos = sImport.length;
        } // switch (sImport[0])
        // kill everything before iPos
        sImport = sImport.substr(iPos);
      } // else need to cope with special chars
    }
    // create clean sImport
    sImport = rImport.join("");
    this.dd("ImportPrefs: clean sImport=\n"+sImport);
    this.dd("ImportPrefs: string table="+osImport.toSource());
    // find anonymous section
    iPos = sImport.search(gksPrefNodePrefix);
    if (iPos < 0)
    { // The anonymous section is the only one existing!
      // Only use it if we have a folder name.
      if (asFolderURI)
        ImportSection(asFolderURI, sImport, osImport);
    }
    else
    { // we have at least one export section
      const kiNodePrefixLen = gksPrefNodePrefix.length;
      var sDefault = sImport.substr(0, iPos); // extract anonymous section
      this.dd("ImportPrefs: sDefault="+sDefault);
      sImport = sImport.substr(iPos);         // and skip it
      var sFolderURI, sSection, iNext;
      while (iPos >= 0)
      { // When (re)entering this loop, sImport has to start with a Mnenhy export section!
        iPos = sImport.search(/\s/);
        sFolderURI = sImport.substring(kiNodePrefixLen, iPos);
        sImport = sImport.substr(iPos);
        // now sImport starts with raw pref data
        iNext = sImport.search(gksPrefNodePrefix); // next section
        sSection = sDefault;
        if (iNext < 0)
        { // no more sections, read until EOF
          sSection += sImport;
        }
        else
        { // read section
          sSection += sImport.substr(0, iNext);
          sImport = sImport.substr(iNext);
        }
        if (!asFolderURI || sFolderURI == asFolderURI)
          ImportSection(sFolderURI, sSection, osImport);
        iPos = iNext;
      }
    }
    goMnenhy.UnsetWaitCursor(window);
  }
}
function ImportSection(asFolderURI, asSection, aoStringTable)
{ // read user_prefs into folderstore.rdf
  this.dd("ImportSection(["+asFolderURI+"],["+asSection+"],["+aoStringTable.toSource()+"])");
  if (!asFolderURI || !asSection)
    return;

  // If this folder has no data of its own yet,
  // create the node, but *not* the threadpane data!
  // (We don't have the window available that
  //  would provide the global data!)
  if (!this.rdf.ExistNode(asFolderURI))
    this.rdf.AddContainer(asFolderURI);
  if (!this.rdf.ExistNode(asFolderURI))
    return;

  const krPrefLines = asSection.match(gkxPrefLines);
  if (!krPrefLines)
    return;

  // have pref lines
  for (var i in krPrefLines)
  { // loop over all pref lines
    const koPrefLine = krPrefLines[i].match(gkxPrefLine);
    if (koPrefLine && koPrefLine.length == 3)
    { // have valid pref line
      var sPref = koPrefLine[1];
      if (sPref in aoStringTable)
      {
        sPref = goMnenhy.ksMnenhyNamespace + encodeURIComponent(aoStringTable[sPref]);
        var sValue = koPrefLine[2];
        if (sValue == "null")
          sValue = null;  // denotes that this pref should be deleted
        else if (sValue == "true" || sValue == "false")
          sValue = this.gkoTypePrefix[goMnenhy.kiPrefBool] + sValue;
        else if (sValue in aoStringTable)
        { // we need to unescape all \\ and \" here
          sValue = this.gkoTypePrefix[goMnenhy.kiPrefString] + aoStringTable[sValue].replace(/\\"/g, '"').replace(/\\\\/g, '\\');
        }
        else
          sValue = this.gkoTypePrefix[goMnenhy.kiPrefInt] + parseInt(sValue);

        if (sPref)
        { // add to folderstore.rdf - a null value means removal from watch list!
          this.dd("ImportSection: set pref: "+sPref+"="+sValue);
          this.rdf.ChangeItemValue(sValue, sPref, asFolderURI);
        }
      } // have pref name
    } // have valid pref line
  } // loop over all pref lines
}



function ExportPrefs(asFolderURI /* = "" */)
{ // Write watched preferences into a user defined file.
  // If no specific asFolderURI is given, loop over all watched preferences.
  // Every set of prefs will be prepended by a comment line with the folder URI.
  this.dd("ExportPrefs("+asFolderURI+")");
  var oFP = Components.classes[goMnenhy.ksCID_FilePicker].createInstance(gkoFilePicker);
  oFP.init(window, gksFilePickerTitleEx, gkoFilePicker.modeSave);
  oFP.appendFilter(gksFilePickerJSTitle,  gksFilePickerJSFile );
  oFP.appendFilter(gksFilePickerAllTitle, gksFilePickerAllFile);
  oFP.displayDirectory = goMnenhy.m_koDirUser.clone();  // default = user profile
  oFP.defaultString    = GetImExFileName(asFolderURI);
  const kiResult = oFP.show();
  if (kiResult == gkoFilePicker.returnOK || kiResult == gkoFilePicker.returnReplace)
  {
    goMnenhy.SetWaitCursor(window);
    var sExport = "";
    // loop over all our containers
    var oContIter;
    if (asFolderURI)
    { // fake nsISimpleEnumerator for single asFolderURI
      oContIter = {uri:             this.rdf.GetRes(asFolderURI),
                   hasMoreElements: function(){return !!this.uri;},
                   getNext:         function(){const o = this.uri; this.uri=null; return o;}
                  };
    }
    else
    {
      oContIter = this.rdf.GetSourcesForNode(goMnenhy.rdf.gksRDFNamespace + goMnenhy.rdf.gksRDFInstanceOf,
                                             goMnenhy.rdf.gksRDFNamespace + goMnenhy.rdf.gksRDFContSeq);
    }
    while (oContIter.hasMoreElements())
    {
      var oCont = oContIter.getNext();
      if (oCont instanceof goMnenhy.rdf.gkIRDFResource)
      { // loop over all outgoing mnenhy arcs that have literal targets
        var oExport = {content: "", fs: this};
        this.rdf.GetAllArcsOut(oCont, this.ExportPrefArc, oExport);
        if (oExport.content)
        { // add prefs to file data
          sExport += gksPrefNodePrefix + oCont.Value + gksNewLine + oExport.content
        }
      }
    } // all containers
    if (sExport)
    {
      this.dd("ExportPrefs:\n" + sExport);
      goMnenhy.WriteFile(oFP.file, sExport);
    }
    goMnenhy.UnsetWaitCursor(window);
  }
}
function ExportPrefArc(aoResourceIgnored, aoItem, aoValue, aoUserData)
{ // add the pref arc to the file content
  if (    (aoItem  instanceof goMnenhy.rdf.gkIRDFResource)
       && (aoValue instanceof goMnenhy.rdf.gkIRDFLiteral))
  {
    const koThis = aoUserData.fs;
    if (!aoItem.Value.search(koThis.gkxTestNS))
    { // create new line in export file
      aoUserData.content += koThis.gksPrefLinePrefix
                            + decodeURIComponent(aoItem.Value.substr(goMnenhy.ksMnenhyNamespace.length))
                            + koThis.gksPrefLineMid;
      const kbString = (aoValue.Value.substr(0, koThis.gkiTypePrefixLen)
                        == koThis.gkoTypePrefix[goMnenhy.kiPrefString]);
      if (kbString)
      { // if we're writing a string, we need to escape all " - and then of course the \ also...
        aoUserData.content += '"' + aoValue.Value.substr(koThis.gkiTypePrefixLen)
                                           .replace(/\\/g, '\\\\')
                                           .replace(/\"/g, '\\"')
                            + '"';
      }
      else
        aoUserData.content += aoValue.Value.substr(koThis.gkiTypePrefixLen);
      aoUserData.content += gksPrefLinePostfix;
    }
  }
  return true; // continue with next element
}


// |
// |  Init called by the global Mnenhy object when loading this script file
// |
function Init(aoMnenhy)
{ // keep pointer to have the Mnenhy object prototype available
  this.goMnenhy = aoMnenhy;
  this.gkxTestNS = new RegExp('^' + goMnenhy.ksMnenhyNamespace);
  // load properties texts
  const koProperties = goMnenhy.InitStringBundle("chrome://mnenhy/locale/mnenhy-folderstore.properties");
  this.gksConfirmInherit     = goMnenhy.GetString("ConfirmInherit",        koProperties);
  this.gksFilePickerTitleIm  = goMnenhy.GetString("FilePickerTitleImport", koProperties);
  this.gksFilePickerTitleEx  = goMnenhy.GetString("FilePickerTitleExport", koProperties);
  this.gksFilePickerJSTitle  = goMnenhy.GetString("FilePickerJSTitle",     koProperties);
  this.gksFilePickerJSFile   = goMnenhy.GetString("FilePickerJSFile",      koProperties);
  this.gksFilePickerAllTitle = goMnenhy.GetString("FilePickerAllTitle",    koProperties);
  this.gksFilePickerAllFile  = goMnenhy.GetString("FilePickerAllFile",     koProperties);
  // init type prefixes
  // these *must* be of equal length!
  this.gkoTypePrefix[goMnenhy.kiPrefString] = "S:";
  this.gkoTypePrefix[goMnenhy.kiPrefInt   ] = "I:";
  this.gkoTypePrefix[goMnenhy.kiPrefBool  ] = "B:";
  this.gkiTypePrefixLen = this.gkoTypePrefix[goMnenhy.kiPrefString].length;
  // init the debug mode
  const kbDebug = goMnenhy.GetPref(this.gksPrefDebug, goMnenhy.kiPrefBool);
  this.dd = kbDebug ? function(asText){goMnenhy.Log(asText),goMnenhy.ConsoleService(asText);}
                    : function(ignored){};
  // get the mnenhy-folderstore.rdf
  var oFolderStore = goMnenhy.m_koDirUser.clone();
  oFolderStore.append(gksFolderstoreRDF);
  const ksFolderStoreURI = goMnenhy.GetProtocolHandler().getURLSpecFromFile(oFolderStore);
  // hack around a trunk (only?) problem that leads to rdf files only consisting of <RDF/>.
  try
  { // delete files that are too short to be correct
    if (!oFolderStore.exists() || (oFolderStore.fileSize < 10))
    { // create empty default storage file
      goMnenhy.WriteFile(oFolderStore,
          '<?xml version="1.0"?>'
        + '<RDF:RDF xmlns:NS1="http://mnenhy.tprac.de/rdf#"'
        +         ' xmlns:NC="http://home.netscape.com/NC-rdf#"'
        +         ' xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">"'
        + '</RDF:RDF>');
    }
  }
  catch (ignored){}
  // now open it
  this.rdf = new goMnenhy.rdf.MnenhyRDF(ksFolderStoreURI, null, null, null, null, null, kbDebug);
  // init the onfolderchange listener for the folder-dependent pref storage
  goMnenhy.m_koObserverService.addObserver({observe: this.MnenhyFolderChange},
                                            this.gksFolderObserver,
                                            false);
}
